function [Estimation]=EstimateTPeel(Samples)
% Estimate the degrees of freedom parameter of a Student-t distribution
% with the method by Peel & McLachlan, "Robust mixture modelling using the t distribution",
% Statistics and Computing (2000) 10, 339348
% Based on equations 16, 20 and 32 of the above mentioned paper.
% Note: Here we do not reevaluate the Mahalanobis distances, which are precomputed
% with respect to the covariance matrix. This is a significant optimization with respect
% to the original version, where the Mahalanobis distances with respect to Sigma are
% recomputed each time the degrees of freedom estimation changes.

Estimation.Means=mean(Samples,2);
Estimation.Covariances=cov(Samples');
while rcond(Estimation.Covariances)<1e-10
    disp('Singular covariance matrix in EstimateTPeel.m')
    Estimation.Covariances=Estimation.Covariances+0.1*eye(size(Estimation.Covariances));
end
InvCovariances=inv(Estimation.Covariances);
[Dimension NumSamples]=size(Samples);
for NdxSample=1:NumSamples
    Difference=Samples(:,NdxSample)-Estimation.Means;
    Estimation.DistMahal(NdxSample)=Difference'*InvCovariances*Difference;
end

% Iterative algorithm to find the degrees of freedom parameter

% Initial estimation
Estimation.DegreesFreedom=5;
ContinueIt=1;
Iterations=1;

while ( ContinueIt && (Iterations<1000) )
    
    % Find the estimated values of the auxiliary random variable w
    Factor=Estimation.DegreesFreedom/(Estimation.DegreesFreedom-2);
    % Mahalanobis distance with respect to Sigma, instead of the covariance
    % matrix C
    Estimation.DistSigma=Factor*Estimation.DistMahal;
    % Estimator for w
    Estimation.W=(Estimation.DegreesFreedom+Dimension)*((Estimation.DegreesFreedom+Estimation.DistSigma).^(-1));

    % Estimation by Peel & McLachlan
    PreviousDegreesFreedom=Estimation.DegreesFreedom;
    Average=mean(log(Estimation.W)-Estimation.W);
    Estimation.DegreesFreedom=fminbnd(@(x) (-psi(0.5*x)+log(0.5*x)+1+...
        Average+psi(0.5*(x+Dimension))-log(0.5*(x+Dimension)))^2,2.1,100);
    
    % Check convergence
    ContinueIt= (abs(Estimation.DegreesFreedom-PreviousDegreesFreedom)/Estimation.DegreesFreedom)>0.001;
    Iterations=Iterations+1;
     
end