function [Estimation]=EstimateShoham(Samples)
% Estimate the degrees of freedom of a Student-t distribution by Shoham's
% mehtod, "Robust clustering by deterministic agglomeration EM of mixtures 
% of multivariate t-distributions",
% Pattern Recognition 35 (2002) 11271142.
% Based on equations 27, 29 and 30 of the above mentioned paper, with beta=1.
% Note: Here we do not reevaluate the Mahalanobis distances, which are precomputed
% with respect to the covariance matrix. This is a significant optimization with respect
% to the original version, where the Mahalanobis distances with respect to Sigma are
% recomputed each time the degrees of freedom estimation changes.


Estimation.Means=mean(Samples,2);
Estimation.Covariances=cov(Samples');
while rcond(Estimation.Covariances)<1e-10
    disp('Singular covariance matrix in EstimateTShoham.m')
    Estimation.Covariances=Estimation.Covariances+0.1*eye(size(Estimation.Covariances));
end
InvCovariances=inv(Estimation.Covariances);
[Dimension NumSamples]=size(Samples);
for NdxSample=1:NumSamples
    Difference=Samples(:,NdxSample)-Estimation.Means;
    Estimation.DistMahal(NdxSample)=Difference'*InvCovariances*Difference;
end

% Iterative algorithm

% Initial estimation
Estimation.DegreesFreedom=5;
ContinueIt=1;
Iterations=1;

while ( ContinueIt && (Iterations<1000) )
    
    % Find the estimated values for the auxiliary random variables l and u
    Factor=Estimation.DegreesFreedom/(Estimation.DegreesFreedom-2);
    % Mahalanobis distance with respect to Sigma, instead of the covariance
    % matrix C
    Estimation.DistSigma=Factor*Estimation.DistMahal;
    % Estimator for l
    Estimation.L=psi(0.5*(Estimation.DegreesFreedom+Dimension))+...
        log(2*((Estimation.DegreesFreedom+Estimation.DistSigma).^(-1)));
    % Estimator for u
    Estimation.U=(Estimation.DegreesFreedom+Dimension)*...
        ((Estimation.DegreesFreedom+Estimation.DistSigma).^(-1));

    % Shoham's estimation
    PreviousDegreesFreedom=Estimation.DegreesFreedom;
    Average=mean(Estimation.L)-mean(Estimation.U);
    Estimation.DegreesFreedom=fminbnd(@(x) (Average+...
        log(0.5*x)+1-psi(0.5*x))^2,2.1,100);
    
    
    % Check convergence
    ContinueIt= (abs(Estimation.DegreesFreedom-PreviousDegreesFreedom)/Estimation.DegreesFreedom)>0.001;
    
    Iterations=Iterations+1;
     
end


