#include "mex.h"
#include "Mates.h"
#include "Debugging.h"

#include "gmp.h"
#include "mpfr.h"
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <float.h>
#include <string.h>

/* 
In order to compile this MEX function, type the following at the MATLAB prompt:
32-bit Windows:
mex TSOMANLLMEX.c MatesLap.c lapack.a blas.a libf2c.a libmpfr.a libgmp.a Debugging.c
64-bit Windows:
mex LINKFLAGS="$LINKFLAGS /NODEFAULTLIB:LIBCMT" TSOMANLLMEX.c MatesLap.c Debugging.c clapack_nowrap.lib BLAS_nowrap.lib libf2c.lib mpir.lib mpfr.lib

Usage:
      
[ANLL,LogDensitiesProb,Respon] = TSOMANLLMEX(Samples,Model);

Samples=Test samples
Model=TSOM model
ANLL=Average negative log-likelihood of the test samples
LogDensitiesProb(n)=Log-density log(p(t sub n)) for all test samples t sub n
Respon(i,n)=Posterior responsibility P( i | t sub n) that unit i has generated test sample t sub n

*/

void FindLogDensTSOMMEX(const mxArray* Model,double *ptrSample,
    double *LogDensityProb,double *ptrRespon,double *ptrLogDetC);

/* Natural logarithm of the gamma function */
double gammln(double argument);


void mexFunction(int nlhs, mxArray* plhs[],
                 int nrhs, const mxArray* prhs[])
{  
    int SpaceDimension,NumSamples,IndexPat,NumRowsMap,NumColsMap,NdxDim,Index;
    int NdxCol,NdxRow;
    const int *DimSamples;
    const int *DimMap;
    double *ptrSamples,*Pattern,*ptrANLL,*ptrLogDensitiesProb,*ptrLogDetC,*ptrMyC;
    double *ptrMyEigenValues,*ptrMyEigenVectors,*ptrRespon;
    double SumLogDensitiesProb;
    mxArray *C,*Nu;

    
    
    /* Get input data */
    DimSamples=mxGetDimensions(prhs[0]);
    SpaceDimension=DimSamples[0];
    NumSamples=DimSamples[1];
    ptrSamples=mxGetPr(prhs[0]);
	Nu=mxGetField(prhs[1],0,"Nu");
    DimMap=mxGetDimensions(Nu);
    NumRowsMap=DimMap[0];
    NumColsMap=DimMap[1];
    C=mxGetField(prhs[1],0,"C");
    
    /* Create output matrices */
    plhs[0]=mxCreateNumericMatrix(1, 1, mxDOUBLE_CLASS, mxREAL);
    ptrANLL=mxGetPr(plhs[0]);
    plhs[1]=mxCreateNumericMatrix(1, NumSamples, mxDOUBLE_CLASS, mxREAL);
    ptrLogDensitiesProb=mxGetPr(plhs[1]);
	plhs[2]=mxCreateNumericMatrix(NumRowsMap*NumColsMap, NumSamples, mxDOUBLE_CLASS, mxREAL);
    ptrRespon=mxGetPr(plhs[2]);
    
    /* Find logarithms of the determinants of the covariance matrices */    
    ptrLogDetC=mxMalloc(NumRowsMap*NumColsMap*sizeof(double));
    ptrMyEigenValues=mxMalloc(SpaceDimension*1*sizeof(double));
    ptrMyEigenVectors=mxMalloc(SpaceDimension*SpaceDimension*sizeof(double));   
    Index=0;
    for(NdxCol=0;NdxCol<NumColsMap;NdxCol++)
    {
        for(NdxRow=0;NdxRow<NumRowsMap;NdxRow++)
        {
            /* Prepare working variables */
            ptrMyC=mxGetPr(mxGetCell(C,Index));
            
            /* Find the eigenvalues for the determinant of the covariance matrix */
            EigValVec(ptrMyC,ptrMyEigenValues,ptrMyEigenVectors,SpaceDimension);    
            
            /* Finding log(det C) */
            ptrLogDetC[Index]=0.0;
            for(NdxDim=0;NdxDim<SpaceDimension;NdxDim++)
            {
                (ptrLogDetC[Index])+=log(ptrMyEigenValues[NdxDim]);
            }
            
            Index++;        
        }
    }
    
    /* Main loop which finds the logarithms of the densities */
    SumLogDensitiesProb=0.0;
    for(IndexPat=0;IndexPat<NumSamples;IndexPat++)
    {
        
        /* Pattern = Samples(:,IndexPat); */
        Pattern=ptrSamples+IndexPat*SpaceDimension;
        
        
        FindLogDensTSOMMEX(prhs[1],Pattern,
                ptrLogDensitiesProb+IndexPat,
				ptrRespon+IndexPat*NumRowsMap*NumColsMap,
				ptrLogDetC);
        SumLogDensitiesProb+=ptrLogDensitiesProb[IndexPat];
        
    }    
    (*ptrANLL)=-SumLogDensitiesProb/(double)NumSamples;  
    
    /* Release memory */
    mxFree(ptrLogDetC); 
    mxFree(ptrMyEigenValues);
    mxFree(ptrMyEigenVectors);   
}    





/*--------------------------------------------------------------------*/

void FindLogDensTSOMMEX(const mxArray* Model,double *ptrSample,
    double *LogDensityProb,double *ptrRespon,double *ptrLogDetC)
{   
    const int *DimMap;
    const int *DimSamples;
    
    int SpaceDimension,NumRowsMap,NumColsMap,Index;
    int NdxCol,NdxRow,NdxDim;
    mxArray *Samples,*Means,*C,*Nu,*Pi,*CInv;
    double *ptrVectorDif,*ptrVectorProd,*ptrMyMean,*ptrMyC,*ptrMyNu,*ptrMiInvC,*ptrPi;
    double *ptrMyCInv;
    double LogDetC,MyLogDensity,D,MyNu,DistMahal;
    mpfr_t SumDensities,LogDens,LL,aux;
	mpfr_t *Dens;
    
    
    mpfr_set_default_prec(100);
    
    /* Get input data */
    Means=mxGetField(Model,0,"Means");
    C=mxGetField(Model,0,"C");
    CInv=mxGetField(Model,0,"CInv");
    Nu=mxGetField(Model,0,"Nu");
    DimMap=mxGetDimensions(Nu);
    NumRowsMap=DimMap[0];
    NumColsMap=DimMap[1];
    Samples=mxGetField(Model,0,"Samples");
    DimSamples=mxGetDimensions(Samples);
    SpaceDimension=DimSamples[0];
    Pi=mxGetField(Model,0,"Pi");
    ptrPi=mxGetPr(Pi);
    
    
    
    /* Create auxiliary matrices */
    ptrVectorDif=mxMalloc(SpaceDimension*1*sizeof(double));
    ptrVectorProd=mxMalloc(SpaceDimension*1*sizeof(double));
    
    
       
    
    mpfr_init(aux);
    mpfr_init(SumDensities); 
    mpfr_set_si(SumDensities,0,GMP_RNDN);
    mpfr_init(LL);
    mpfr_init(LogDens);   
    Dens=mxMalloc(NumRowsMap*NumColsMap*sizeof(mpfr_t));
	for(Index=0;Index<NumRowsMap*NumColsMap;Index++)
	{
		mpfr_init(Dens[Index]); 
	}
    
    Index=0;
    for(NdxCol=0;NdxCol<NumColsMap;NdxCol++)
    {
        for(NdxRow=0;NdxRow<NumRowsMap;NdxRow++)
        {
            /* Prepare working variables */
            ptrMyMean=mxGetPr(mxGetCell(Means,Index));
            ptrMyC=mxGetPr(mxGetCell(C,Index));
            ptrMyCInv=mxGetPr(mxGetCell(CInv,Index));            
            ptrMyNu=mxGetPr(mxGetCell(Nu,Index));
            
            
            /* VectorDif=Samples(:,NdxSample) - Model.Mu{NdxRow,NdxCol}; */
            MatrixDifference(ptrSample,ptrMyMean,
                    ptrVectorDif,SpaceDimension,1);
            
            /* Find squared Mahalanobis distance with respect to the covariance matrix */
            MatrixProduct(ptrMyCInv,ptrVectorDif,ptrVectorProd,SpaceDimension,
                    SpaceDimension,1);
            MatrixProduct(ptrVectorDif,ptrVectorProd,&DistMahal,
                    1,SpaceDimension,1);
            
            
                                                         
            /* Finding log(p(t sub n | i) */
            D=(double)SpaceDimension;
            MyNu=(*ptrMyNu);
            
            MyLogDensity=gammln(0.5*D+0.5*MyNu)-gammln(0.5*MyNu)-0.5*ptrLogDetC[Index]
                 -0.5*D*(1.144729885849400+log(MyNu-2.0))
                 -0.5*(D+MyNu)*log(1.0+DistMahal/(MyNu-2.0));
            
            /* This is p(t sub n | i) */
            /*ptrDensities[Index]=exp(MyLogDensity);*/
            mpfr_set_d(LogDens,MyLogDensity,GMP_RNDN);
            mpfr_exp (Dens[Index], LogDens, GMP_RNDN); /* Exponential e ^ datum */
           /* Find (pi sub i) * p(t sub n | i) in the variable Dens and add it in order to 
            compute p(t sub n) in the variable SumDensities */  
            mpfr_set_d(aux,ptrPi[Index],GMP_RNDN);
            mpfr_mul(Dens[Index],Dens[Index],aux,GMP_RNDN);
            mpfr_add(SumDensities,SumDensities,Dens[Index], GMP_RNDN);
            
            /* Index to iterate through the cells */
            Index++;
     
        }    
 
    }    
    
    /* Find log( p( t sub n ) ) in the variable LL */
    mpfr_log(LL, SumDensities, GMP_RNDN);
    (*LogDensityProb)=mpfr_get_d(LL, GMP_RNDN);

	/* Find responsibilities P( i | t sub n) */
	for(Index=0;Index<NumRowsMap*NumColsMap;Index++)
	{
		mpfr_div(Dens[Index],Dens[Index],SumDensities, GMP_RNDN);
		ptrRespon[Index]=mpfr_get_d(Dens[Index], GMP_RNDN);
	}
    
    
    mpfr_clear(SumDensities); /* Release memory */     
    mpfr_clear(LogDens); /* Release memory */        
    mpfr_clear(LL); /* Release memory */
    mpfr_clear(aux); /* Release memory */
	for(Index=0;Index<NumRowsMap*NumColsMap;Index++)
	{
		mpfr_clear(Dens[Index]); 
	}
	mxFree(Dens);
    
    mxFree(ptrVectorDif);
    mxFree(ptrVectorProd);

    
    
}    


/*

  Purpose:

    ALNGAM computes the logarithm of the gamma function.

  Modified:

    13 January 2008

  Author:

    Allan Macleod
    C++ version by John Burkardt

  Reference:

    Allan Macleod,
    Algorithm AS 245,
    A Robust and Reliable Algorithm for the Logarithm of the Gamma Function,
    Applied Statistics,
    Volume 38, Number 2, 1989, pages 397-402.

  Parameters:

    Input, double XVALUE, the argument of the Gamma function.

    Output, int IFAULT, error flag.
    0, no error occurred.
    1, XVALUE is less than or equal to 0.
    2, XVALUE is too big.

    Output, double ALNGAM, the logarithm of the gamma function of X.
*/
double alngam ( double xvalue, int *ifault )
{
  double alr2pi = 0.918938533204673;
  double r1[9] = {
    -2.66685511495, 
    -24.4387534237, 
    -21.9698958928, 
     11.1667541262, 
     3.13060547623, 
     0.607771387771, 
     11.9400905721, 
     31.4690115749, 
     15.2346874070 };
  double r2[9] = {
    -78.3359299449, 
    -142.046296688, 
     137.519416416, 
     78.6994924154, 
     4.16438922228, 
     47.0668766060, 
     313.399215894, 
     263.505074721, 
     43.3400022514 };
  double r3[9] = {
    -2.12159572323E+05, 
     2.30661510616E+05, 
     2.74647644705E+04, 
    -4.02621119975E+04, 
    -2.29660729780E+03, 
    -1.16328495004E+05, 
    -1.46025937511E+05, 
    -2.42357409629E+04, 
    -5.70691009324E+02 };
  double r4[5] = {
     0.279195317918525, 
     0.4917317610505968, 
     0.0692910599291889, 
     3.350343815022304, 
     6.012459259764103 };
  double value;
  double x;
  double x1;
  double x2;
  double xlge = 510000.0;
  double xlgst = 1.0E+30;
  double y;

  x = xvalue;
  value = 0.0;
/*
  Check the input.
*/
  if ( xlgst <= x )
  {
    *ifault = 2;
    return value;
  }

  if ( x <= 0.0 )
  {
    *ifault = 1;
    return value;
  }

  *ifault = 0;
/*
  Calculation for 0 < X < 0.5 and 0.5 <= X < 1.5 combined.
*/
  if ( x < 1.5 )
  {
    if ( x < 0.5 )
    {
      value = - log ( x );
      y = x + 1.0;
/*
  Test whether X < machine epsilon.
*/
      if ( y == 1.0 )
      {
        return value;
      }
    }
    else
    {
      value = 0.0;
      y = x;
      x = ( x - 0.5 ) - 0.5;
    }

    value = value + x * (((( 
        r1[4]   * y 
      + r1[3] ) * y 
      + r1[2] ) * y 
      + r1[1] ) * y 
      + r1[0] ) / (((( 
                  y 
      + r1[8] ) * y 
      + r1[7] ) * y 
      + r1[6] ) * y 
      + r1[5] );

    return value;
  }
/*
  Calculation for 1.5 <= X < 4.0.
*/
  if ( x < 4.0 )
  {
    y = ( x - 1.0 ) - 1.0;

    value = y * (((( 
        r2[4]   * x 
      + r2[3] ) * x 
      + r2[2] ) * x 
      + r2[1] ) * x 
      + r2[0] ) / (((( 
                  x 
      + r2[8] ) * x 
      + r2[7] ) * x 
      + r2[6] ) * x 
      + r2[5] );
  }
/*
  Calculation for 4.0 <= X < 12.0.
*/
  else if ( x < 12.0 ) 
  {
    value = (((( 
        r3[4]   * x 
      + r3[3] ) * x 
      + r3[2] ) * x 
      + r3[1] ) * x 
      + r3[0] ) / (((( 
                  x 
      + r3[8] ) * x 
      + r3[7] ) * x 
      + r3[6] ) * x 
      + r3[5] );
  }
/*
  Calculation for 12.0 <= X.
*/
  else
  {
    y = log ( x );
    value = x * ( y - 1.0 ) - 0.5 * y + alr2pi;

    if ( x <= xlge )
    {
      x1 = 1.0 / x;
      x2 = x1 * x1;

      value = value + x1 * ( ( 
             r4[2]   * 
        x2 + r4[1] ) * 
        x2 + r4[0] ) / ( ( 
        x2 + r4[4] ) * 
        x2 + r4[3] );
    }
  }

  return value;
}

/* Natural logarithm of the gamma function */
double gammln(double argument)
{
	int ifault;
    
    return alngam(argument,&ifault);

}



