function [Model]= TrainTSOM(Samples,NumRowsMap,NumColsMap,Method,OutputsTimeSeries,NumSteps)
% Inputs
%   Samples: Matrix [Dimension,NumSamples] with the training data, or the
%   name of the function that provides the samples (string)
%   NumRowsMap,NumColsMap: size of the self organizing map
%   NumSteps: Number of steps of the training phase
%   OutputsTimeSeries: If the samples are the sequential values of a time series
%       that we want to study, this is a matrix [1,NumSamples] with the 
%       corresponding desired outputs. Otherwise, it is the empty matrix.
%   Method: Method to estimate the degrees of freedom parameter (nu)
%       Method=0 for Peel => AverageNu=E[log(u)-u]
%       Method=1 for Shoham => AverageNu=E[l-u]
%       Method=2 for the closed-form method => AverageNu=E[log(sqrt(rho))]

if ischar(Samples)
    Result= feval(Samples,0);
    Dimension=Result(1);
    NumSamples=Result(2);
else
    [Dimension, NumSamples] = size(Samples);
end

%   Structure initialization

Model.Samples=Samples;
Model.NumRowsMap=NumRowsMap;
Model.NumColsMap=NumColsMap;
Model.Method=Method;
Model.TimeSeries=double(~isempty(OutputsTimeSeries));
if (Model.TimeSeries)
    Model.OutputsTimeSeries=OutputsTimeSeries;
end

%   We choose some samples at random to initialize each neuron 
rand('state',sum(100*clock));
NumNeuronas=NumRowsMap*NumColsMap;
NumPatIni = floor(NumSamples / NumNeuronas);
fprintf('Initializing neurons')
for NdxRow=1:NumRowsMap
    for NdxCol=1:NumColsMap
        IndicePatIni = round(rand(1,NumPatIni)*(NumSamples-1))+1;
        if ischar(Samples)
            PatIni=feval(Samples,IndicePatIni);
        else
            PatIni = Samples(:,IndicePatIni);
        end
        
        switch Method
            case 0
                % Peel
                Estimation=EstimateTPeel(PatIni);
                Model.Means{NdxRow,NdxCol}=Estimation.Means;
                Model.C{NdxRow,NdxCol}=Estimation.Covariances;
                Model.CInv{NdxRow,NdxCol}=inv(Model.C{NdxRow,NdxCol});
                Model.Nu{NdxRow,NdxCol}=Estimation.DegreesFreedom;
                Model.AverageNu{NdxRow,NdxCol}=...
                    mean(log(Estimation.W)-Estimation.W);
            case 1
                % Shoham
                Estimation=EstimateTShoham(PatIni);
                Model.Means{NdxRow,NdxCol}=Estimation.Means;
                Model.C{NdxRow,NdxCol}=Estimation.Covariances;
                Model.CInv{NdxRow,NdxCol}=inv(Model.C{NdxRow,NdxCol});
                Model.Nu{NdxRow,NdxCol}=Estimation.DegreesFreedom;
                Model.AverageNu{NdxRow,NdxCol}=...
                    mean(Estimation.L)-mean(Estimation.U);
            case 2
                % Closed
                Estimation=EstimateTClosed(PatIni);
                Model.Means{NdxRow,NdxCol}=Estimation.Means;
                Model.C{NdxRow,NdxCol}=Estimation.Covariances;
                Model.CInv{NdxRow,NdxCol}=inv(Model.C{NdxRow,NdxCol});
                Model.Nu{NdxRow,NdxCol}=Estimation.DegreesFreedom;
                Model.AverageNu{NdxRow,NdxCol}=...
                    mean(Estimation.Rho);
            otherwise
                error('Unknown method')
        end
        
        if Model.TimeSeries==1
            Model.Coefficients{NdxRow,NdxCol}=(1/Dimension)*ones(Dimension,1);
        end
                      
            
        fprintf('.')
    end
end

Model.NumVictories=zeros(NumRowsMap,NumColsMap);
%   Equal a priori probabilities
Model.Pi=(1/(NumRowsMap*NumColsMap))*ones(NumRowsMap,NumColsMap);
fprintf('\nThe neurons are initialized\n')
fprintf('Training the map\n')
clear PatIni;

%----------------------------------------------------
% Training phase
%----------------------------------------------------


Model = TrainTSOMMEX(Model,NumSteps,Method);



fprintf('Training finished\n')
        
                
