#include "mex.h"
#include "Mates.h"


#include "gmp.h"
#include "mpfr.h"
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <float.h>
#include <string.h>

#define METHOD_PEEL 0
#define METHOD_SHOHAM 1
#define METHOD_CLOSED 2
#define ALPHA_PRIME 0.001


/* 

In order to compile this MEX function, type the following at the MATLAB prompt:
32-bit Windows:
mex TrainTSOMMEX.c MatesLap.c lapack.a blas.a libf2c.a libmpfr.a libgmp.a Debugging.c
64-bit Windows:
mex LINKFLAGS="$LINKFLAGS /NODEFAULTLIB:LIBCMT" TrainTSOMMEX.c MatesLap.c Debugging.c clapack_nowrap.lib BLAS_nowrap.lib libf2c.lib mpir.lib mpfr.lib

Usage:
[Model] = TrainTSOMMEX(Model,NumSteps,Method);

where
Method=0 for Peel => AverageNu=E[log(u)-u]
Method=1 for Shoham => AverageNu=E[l-u]
Method=2 for the closed form method => AverageNu=E[log(sqrt(rho))]


*/

#ifndef isnan
int isnan(double x) {
    return x != x;
}
#endif

#ifndef isinf
int isinf(double x) {
    return ((x - x) != 0);
}
#endif

#ifndef M_PI
#define M_PI 3.141592653589793
#endif

void WinningTSOMProbMEX(mxArray* Model,double *ptrSample,double RadiusNeighbourhood,
    int *NdxWinRow,int *NdxWinCol,double *LogDensityProb,
    double *ptrResponsibilities,double *ptrDistMahal);
    
double psi(double x);
double trigamma( double x );
double gammln(double xx);

/* Solve a nonlinear equation in nu*/
void SolveNuEquation(const mxArray* prhs[],double DistMahal,double D,
       double OldCoef,double NewCoef,double *ptrMyNu,double *ptrMyAverageNu); 

/* Find an interval which contains a solution by splitting the interval [Start,End]
   in NumSlices subintervals */
int FindSubinterval(double (*NonlinearFunction)(double,double,double,double,double,double), double NewAverage, double D,
     double DistMahal,double OldCoef,double NewCoef,
     double Start, double End, int NumSlices,double *ResultStart, double *ResultEnd);

/* Find a solution of an equation by Newton-Raphson and bisection. Needs the derivative */
double FindSolution(double (*func)(double,double,double,double,double,double),
       double (*funcDeriv)(double,double,double,double,double,double),
       double NPr, double D,double DistMahal,double OldCoef,double NewCoef,
       const double Start, const double End,const double Accuracy);


/* Nonlinear function for Peel's method */
double FunctionPeel(double Nu,double NewAverage,double D,double DistMahal,
       double OldCoef,double NewCoef);

/* Derivative of the nonlinear function for Peel's method */
double DerivFunctionPeel(double Nu,double NewAverage,double D,double DistMahal,
       double OldCoef,double NewCoef);
       
/* Nonlinear function for Shoham's method */
double FunctionShoham(double Nu,double NewAverage,double D,double DistMahal,
       double OldCoef,double NewCoef);

/* Derivative of the nonlinear function for Shoham's method */
double DerivFunctionShoham(double Nu,double NewAverage,double D,double DistMahal,
       double OldCoef,double NewCoef);
       
/* Nonlinear function for the closed-form method */
double FunctionClosed(double Nu,double NewAverage,double D,double DistMahal,
       double OldCoef,double NewCoef);

/* Derivative of the nonlinear function for the closed-form method */
double DerivFunctionClosed(double Nu,double NewAverage,double D,double DistMahal,
       double OldCoef,double NewCoef);


void mexFunction(int nlhs, mxArray* plhs[],
                 int nrhs, const mxArray* prhs[])
{  
    int SpaceDimension,NumSamples,NumRowsMap,NumColsMap;
    mxArray *Means,*NumVictories;
    mxArray *Samples,*Pi,*C,*Nu,*CInv,*Coefficients;
    mxArray *AverageNu;
    unsigned int Seed;
    int RandomNumber;
    double MaxRadiusNeighbourhood,RadiusNeighbourhood,LogDensityProb;
    double LearningRate,OneMinusRate;
    double *ptrNumVictories,*Pattern,*ptrNumSteps,*ptrMyCoefficients;
    int NdxCol,NdxRow,NdxWinRow,NdxWinCol,NumSteps,NdxStep,IndexPat,NdxDatum;
    double *ptrSamples,*ptrDistMahal;
    const int *DimMap;
    const int *DimSamples;
    double *ptrMyMean,*ptrMyC,*ptrMyNu,*ptrResponsibilities,*ptrPi,*ptrDevMean,*ptrMyCInv;
    double *ptrDevDevT,*ptrMyAverageNu,*ptrMethod,*ptrInvR,*ptrOutputsTimeSeries;
    double NewCoef,OldCoef,AntPi,SumPis,D,OutputDesired,OutputObtained,SquaredNormPattern;
    int Index,Method,TimeSeries;
    double NormC,NormInvC,ConditionNumber;
 
    /* Get input data */
    ptrNumSteps=mxGetPr(prhs[1]);
    NumSteps=(int)(*ptrNumSteps);
    
    NumSteps=labs(NumSteps);
    ptrMethod=mxGetPr(prhs[2]);
    Method=(int)(*ptrMethod); 
    
    /* Setup pseudorandom number generator */
    Seed=time(NULL);
    srand(Seed);
    
    /* Create output array */
    plhs[0]=mxDuplicateArray(prhs[0]);
    
    /* See whether we are studying a time series */ 
    TimeSeries=(int)(*mxGetPr(mxGetField(plhs[0],0,"TimeSeries")));
    if (TimeSeries)
    {
         Coefficients=mxGetField(plhs[0],0,"Coefficients");
         ptrOutputsTimeSeries=mxGetPr(mxGetField(plhs[0],0,"OutputsTimeSeries"));
         
    }
    
    /* Get working variables */
    Means=mxGetField(plhs[0],0,"Means");
    C=mxGetField(plhs[0],0,"C");
    CInv=mxGetField(plhs[0],0,"CInv");
    Nu=mxGetField(plhs[0],0,"Nu");
    AverageNu=mxGetField(plhs[0],0,"AverageNu");
    Pi=mxGetField(plhs[0],0,"Pi");
    ptrPi=mxGetPr(Pi);
    
    DimMap=mxGetDimensions(Nu);
    NumRowsMap=DimMap[0];
    NumColsMap=DimMap[1];
    Samples=mxGetField(plhs[0],0,"Samples");
    DimSamples=mxGetDimensions(Samples);
    SpaceDimension=DimSamples[0];
    D=(double)SpaceDimension;
    NumSamples=DimSamples[1];
    ptrSamples=mxGetPr(Samples);
    NumVictories=mxGetField(plhs[0],0,"NumVictories");
    ptrNumVictories=mxGetPr(NumVictories);
         
    /* Prepare auxiliary arrays */
    ptrDevMean=mxMalloc(SpaceDimension*1*sizeof(double));
    ptrResponsibilities=mxMalloc(NumRowsMap*NumColsMap*sizeof(double));
    ptrDistMahal=mxMalloc(NumRowsMap*NumColsMap*sizeof(double));
    ptrDevDevT=mxMalloc(SpaceDimension*SpaceDimension*sizeof(double));
    ptrInvR=mxMalloc(SpaceDimension*SpaceDimension*sizeof(double));
    
    /* MaxRadiusNeighbourhood=mean([NumRowsMap NumColsMap])/4; */
    MaxRadiusNeighbourhood=((double)NumRowsMap+(double)NumColsMap)/8.0;
    
    /* Main loop */
    for(NdxStep=0;NdxStep<NumSteps;NdxStep++)
    {

        /* Choose a pattern */
        /*IndexPat = round(rand(1)*(NumSamples-1))+1; */
        RandomNumber=rand();
        
        IndexPat=RandomNumber%NumSamples;
        
        /* Pattern = Samples(:,IndexPat); */
        Pattern=ptrSamples+IndexPat*SpaceDimension;
        
        RadiusNeighbourhood=MaxRadiusNeighbourhood*(1.0-(double)NdxStep/NumSteps);
        
        /* Find the winning distribution, the responsibilities and the
           Mahalanobis distances with respect to the covariance matrix */
        WinningTSOMProbMEX(plhs[0],Pattern,RadiusNeighbourhood,
                &NdxWinRow,&NdxWinCol,&LogDensityProb,
                ptrResponsibilities,ptrDistMahal);
                 
        /* Increment Model.NumVictories(NdxWinRow,NdxWinCol) */
        ptrNumVictories[NdxWinRow+NumRowsMap*NdxWinCol]++;
                
        Index=0;
        for(NdxCol=0;NdxCol<NumColsMap;NdxCol++)
        {
            for(NdxRow=0;NdxRow<NumRowsMap;NdxRow++)
            {
                /* Get pointers to this neuron */
                ptrMyMean=mxGetPr(mxGetCell(Means,Index));
                ptrMyC=mxGetPr(mxGetCell(C,Index));
                ptrMyCInv=mxGetPr(mxGetCell(CInv,Index));
                ptrMyNu=mxGetPr(mxGetCell(Nu,Index));
                ptrMyAverageNu=mxGetPr(mxGetCell(AverageNu,Index));
                
                
                /* Find the update rate of this neuron from its responsibility */
                
                /* Values of LearningRate larger than 0.8 may hang this function 
                because either NewCoef or OldCoef become 1 for all neurons,
                and this makes the algorithm converge very slowly.
                On the other hand, we always have OldCoef+NewCoef==1.0 */
                /* Linear decay also works, but it is less justifiable from a mathematical
                point of view:
                LearningRate=0.01*(1.0-(double)(NdxStep+1.0)/(NumSteps+1.0));*/
                LearningRate=1.0/(0.01*(double)NdxStep+100.0);
                OneMinusRate=1.0-LearningRate;
                AntPi=ptrPi[Index];
                ptrPi[Index]=OneMinusRate*ptrPi[Index]+
                                LearningRate*ptrResponsibilities[Index];
                OldCoef=(OneMinusRate*AntPi)/ptrPi[Index];
                NewCoef=(LearningRate*ptrResponsibilities[Index])/ptrPi[Index];
                /* If there are problems, leave things as they were */
                if (isnan(OldCoef) || isinf(OldCoef) || isnan(NewCoef) || isinf(NewCoef) )
                {
                     OldCoef=1.0;
                     NewCoef=0.0;
                }
                
           
                
                /* Model.Means{NdxRow,NdxCol}=...
                    OldCoef*Model.Means{NdxRow,NdxCol}+...
                     NewCoef*Pattern; */
                for(NdxDatum=0;NdxDatum<SpaceDimension;NdxDatum++)
                {
                    ptrMyMean[NdxDatum]=OldCoef*ptrMyMean[NdxDatum]+NewCoef*Pattern[NdxDatum];
                }
                
                
                /* DevMean = Pattern - Model.Means{NdxRow,NdxCol}; */
                         
                MatrixDifference(Pattern,ptrMyMean,ptrDevMean,SpaceDimension,1);
                
                /* DevDevT=DevMean*DevMean'; */
                MatrixProduct(ptrDevMean,ptrDevMean,ptrDevDevT,SpaceDimension,1,SpaceDimension);
                
                /* Model.C{NdxRow,NdxCol}=...
                    OldCoef*Model.C{NdxRow,NdxCol}+...
                     NewCoef*Pattern; */
                ScalarMatrixProduct(NewCoef,ptrDevDevT,ptrDevDevT,SpaceDimension,
                      SpaceDimension);
                ScalarMatrixProduct(OldCoef,ptrMyC,ptrMyC,SpaceDimension,
                      SpaceDimension);
                MatrixSum(ptrMyC,ptrDevDevT,ptrMyC,SpaceDimension,SpaceDimension);
                
                /* Invert the covariance matrix, with previous regularization if needed */
                InverseNorm(ptrMyC,NULL,&NormC,&NormInvC,SpaceDimension,SpaceDimension);
                ConditionNumber=NormC*NormInvC;
                while (ConditionNumber>1.0e6)
                {
                      /* Increase the diagonal elements to avoid matrix singularity */
                      for(NdxDatum=0;NdxDatum<SpaceDimension;NdxDatum++)
                      {
                           ptrMyC[NdxDatum+SpaceDimension*NdxDatum]+=0.1;
                      }
                      InverseNorm(ptrMyC,NULL,&NormC,&NormInvC,SpaceDimension,SpaceDimension);
                      ConditionNumber=NormC*NormInvC;                      
                }
                Inverse(ptrMyC,ptrMyCInv,SpaceDimension);
            
                /* Estimate the number of degrees of freedom */
                /* Solve the nonlinear equation in nu */
                SolveNuEquation(prhs,ptrDistMahal[Index],
                    D,OldCoef,NewCoef,ptrMyNu,ptrMyAverageNu);
                    
                /* If we are studying a time series, update the coefficients
                of the adaptive filter of this neuron */
                if (TimeSeries)
                {
                     ptrMyCoefficients=mxGetPr(mxGetCell(Coefficients,Index));
                     OutputDesired=ptrOutputsTimeSeries[IndexPat];
                     
                     /* Find the correlation matrix in ptrDevDevT */
                     MatrixProduct(ptrMyMean,ptrMyMean,ptrDevDevT,SpaceDimension,1,SpaceDimension);
                     MatrixSum(ptrMyC,ptrDevDevT,ptrDevDevT,SpaceDimension,SpaceDimension);
                     
                     /* Find the inverse of the correlation matrix in ptrInvR */
                     /* with previous regularization if needed */
                     InverseNorm(ptrDevDevT,NULL,&NormC,&NormInvC,SpaceDimension,SpaceDimension);
                     ConditionNumber=NormC*NormInvC;
                     while (ConditionNumber>1.0e6)
                     {
                          /* Increase the diagonal elements to avoid matrix singularity */
                          for(NdxDatum=0;NdxDatum<SpaceDimension;NdxDatum++)
                          {
                               ptrDevDevT[NdxDatum+SpaceDimension*NdxDatum]+=0.1;
                          }
                          InverseNorm(ptrDevDevT,NULL,&NormC,&NormInvC,SpaceDimension,SpaceDimension);
                          ConditionNumber=NormC*NormInvC;                      
                     }
                     Inverse(ptrDevDevT,ptrInvR,SpaceDimension);
                     
                     /* Find inv(R)*Pattern in ptrDevMean */
                     MatrixProduct(ptrInvR,Pattern,ptrDevMean,SpaceDimension,SpaceDimension,1);
                     
                     /* Find the output of the filter */
                     MatrixProduct(ptrMyCoefficients,Pattern,&OutputObtained,1,SpaceDimension,1);
                     
                     /* Find the squared Euclidean norm of Pattern */
                     SquaredNorm(Pattern,&SquaredNormPattern,SpaceDimension);
                     
                     /* Find step_size * inv(R) * e * Pattern in ptrDevMean */
                     ScalarMatrixProduct(ALPHA_PRIME*ptrResponsibilities[Index]*
                          (OutputDesired-OutputObtained),
                          ptrDevMean,ptrDevMean,SpaceDimension,1);
                          
                     /* Add the obtained vector to the previous coefficient vector */
                     MatrixSum(ptrMyCoefficients,ptrDevMean,ptrMyCoefficients,SpaceDimension,1);
                
                }
                  
                /* Index to iterate through the cells */
                Index++;
            }
        }
    }        
    
    
    /* ptrPi[Index] is proportional to the a priori probability of i, P(i). 
    Now we must normalize so that the sum is one */
    SumPis=0.0;
    for(Index=0;Index<NumRowsMap*NumColsMap;Index++)
    {
        SumPis+=ptrPi[Index];
    }    
    for(Index=0;Index<NumRowsMap*NumColsMap;Index++)
    {
        ptrPi[Index]/=SumPis;
    }    

    /* Free auxiliary arrays */
    mxFree(ptrDevMean);
    mxFree(ptrResponsibilities);
    mxFree(ptrDistMahal);
    mxFree(ptrDevDevT);
    mxFree(ptrInvR);
    
}    







/*--------------------------------------------------------------------*/
        
void WinningTSOMProbMEX(mxArray* Model,double *ptrSample,double RadiusNeighbourhood,
    int *NdxWinRow,int *NdxWinCol,double *LogDensityProb,double *ptrResponsibilities,
    double *ptrDistMahal)
{   
    
    const int *DimMap;
    const int *DimSamples;
    
    int SpaceDimension,NumRowsMap,NumColsMap,Index;
    int NdxCol,NdxRow,NdxColDist,NdxRowDist,NdxDim;
    mxArray *Samples,*Means,*C,*Nu,*CInv;
    double *ptrVectorDif,*ptrVectorProd,*ptrMyMean,*ptrMyC,*ptrMyNu,*ptrMyCInv;
    double *ptrMyEigenValues,*ptrMyEigenVectors;
    mpfr_t *ptrDensities;
    double LogDetC,MyLogDensity,SumNeighbourhoodFunctions,D,MyNu;
    double TopologicDistance,*NeighbourhoodFunction;
    mpfr_t DistributionProbDens,MySumNeighbourhoodFunctions,aux;
    
    
    mpfr_set_default_prec(100);
    
    /* Get input data*/
    Means=mxGetField(Model,0,"Means");
    C=mxGetField(Model,0,"C");
    CInv=mxGetField(Model,0,"CInv");
    Nu=mxGetField(Model,0,"Nu");
    DimMap=mxGetDimensions(Nu);
    NumRowsMap=DimMap[0];
    NumColsMap=DimMap[1];
    Samples=mxGetField(Model,0,"Samples");
    DimSamples=mxGetDimensions(Samples);
    SpaceDimension=DimSamples[0];
    
    
    
    /* Create auxiliary arrays */
    ptrVectorDif=mxMalloc(SpaceDimension*1*sizeof(double));
    ptrVectorProd=mxMalloc(SpaceDimension*1*sizeof(double));
    NeighbourhoodFunction=mxMalloc(NumColsMap*NumRowsMap*sizeof(double));
    ptrMyEigenValues=mxMalloc(SpaceDimension*1*sizeof(double));
    ptrMyEigenVectors=mxMalloc(SpaceDimension*SpaceDimension*sizeof(double));
    
    
    ptrDensities=mxMalloc(NumColsMap*NumRowsMap*sizeof(mpfr_t));
    for(Index=0;Index<NumColsMap*NumRowsMap;Index++)
    {
        mpfr_init(ptrDensities[Index]);        
    }     
    mpfr_init(DistributionProbDens);
    mpfr_init(aux);
    mpfr_init(MySumNeighbourhoodFunctions);

    
    Index=0;
    /* LogDensityProb = -inf; */
    (*LogDensityProb)=-DBL_MAX;
    
    for(NdxCol=0;NdxCol<NumColsMap;NdxCol++)
    {
        for(NdxRow=0;NdxRow<NumRowsMap;NdxRow++)
        {
            /* Get working variables */
            ptrMyMean=mxGetPr(mxGetCell(Means,Index));
            ptrMyC=mxGetPr(mxGetCell(C,Index));
            ptrMyCInv=mxGetPr(mxGetCell(CInv,Index));
            ptrMyNu=mxGetPr(mxGetCell(Nu,Index));
            
            
            /* VectorDif=Samples(:,NdxSample) - Model.Mu{NdxRow,NdxCol}; */
            MatrixDifference(ptrSample,ptrMyMean,
                    ptrVectorDif,SpaceDimension,1);
            
            
            /* Find squared Mahalanobis distance with respect to the covariance matrix */
            MatrixProduct(ptrMyCInv,ptrVectorDif,ptrVectorProd,SpaceDimension,
                    SpaceDimension,1);
            MatrixProduct(ptrVectorDif,ptrVectorProd,ptrDistMahal+Index,
                    1,SpaceDimension,1);
            
            /* Find the eigenvalues in order to compute the determinant of the covariance matrix*/
            EigValVec(ptrMyC,ptrMyEigenValues,ptrMyEigenVectors,SpaceDimension);    
            
            /* Finding log(det C) */
            LogDetC=0.0;
            for(NdxDim=0;NdxDim<SpaceDimension;NdxDim++)
            {
                LogDetC+=log(ptrMyEigenValues[NdxDim]);
            }
                                                         
            /* Finding log(p(t sub n | i) */
            D=(double)SpaceDimension;
            MyNu=(*ptrMyNu);
            
            MyLogDensity=gammln(0.5*D+0.5*MyNu)-gammln(0.5*MyNu)-0.5*LogDetC
                 -0.5*D*(1.144729885849400+log(MyNu-2.0))
                 -0.5*(D+MyNu)*log(1.0+ptrDistMahal[Index]/(MyNu-2.0));
            
            /* This is p(t sub n | i) */
            /*ptrDensities[Index]=exp(MyLogDensity);*/
            mpfr_set_d(ptrDensities[Index],MyLogDensity,GMP_RNDN);
            mpfr_exp(ptrDensities[Index], ptrDensities[Index], GMP_RNDN); /* Exponential e ^ datum */
            
            
            /* Index to iterate through the cells */
            Index++;
     
        }    
 
    }    
    
    
    /* LogDensityProb = -inf; */
    (*LogDensityProb)=-DBL_MAX;
    
    /* Compute the probability density of each distribution, p(t sub n | q sub nj) */
    
    for(NdxColDist=0;NdxColDist<NumColsMap;NdxColDist++)
    {
        for(NdxRowDist=0;NdxRowDist<NumRowsMap;NdxRowDist++)
        {
            /* For each neuron in the map */
            Index=0;
            /*DistributionProbDens=0.0;*/
            mpfr_set_si(DistributionProbDens,0,GMP_RNDN);
            SumNeighbourhoodFunctions=0.0;
            for(NdxCol=0;NdxCol<NumColsMap;NdxCol++)
            {
                for(NdxRow=0;NdxRow<NumRowsMap;NdxRow++)
                {
                
                /* Manhattan distance:
                TopologicDistance=abs(NdxRow-NdxWinRow)+abs(NdxCol-NdxWinCol); */
                TopologicDistance=(double)(abs(NdxRow-NdxRowDist)+abs(NdxCol-NdxColDist));
                
                /* NeighbourhoodFunction=exp(-0.5*((TopologicDistance/RadiusNeighbourhood)^2)); */
                /* This is q sub nji (unnormalized) */
                NeighbourhoodFunction[Index]=exp(-0.5*((TopologicDistance*TopologicDistance)
                                /(RadiusNeighbourhood*RadiusNeighbourhood)));
                
                /* Computing (q sub nji) * p(t sub n | i)
                 with unnormalized q sub nji */
                /*aux=NeighbourhoodFunction*ptrDensities[Index];*/
                mpfr_set_d(aux,NeighbourhoodFunction[Index],GMP_RNDN);
                mpfr_mul(aux,aux,ptrDensities[Index],GMP_RNDN);
                
                
                /* Computing p(t sub n | q sub nj)
                 with unnormalized q sub nji*/
                /*DistributionProbDens+=ptrMysResponsibilities[Index];*/
                mpfr_add(DistributionProbDens,DistributionProbDens,
                                aux,GMP_RNDN);
                          
                SumNeighbourhoodFunctions+=NeighbourhoodFunction[Index];
                
                Index++;                
                }    
            }
            
            /* Finding p(t sub n | q sub nj)
             by normalizing the previous value */
            /*DistributionProbDens/=SumNeighbourhoodFunctions;*/
            mpfr_set_d(MySumNeighbourhoodFunctions,SumNeighbourhoodFunctions,GMP_RNDN);
            mpfr_div(DistributionProbDens,DistributionProbDens,MySumNeighbourhoodFunctions,GMP_RNDN);
            
            
            /* See whether this distribution q sub nj yields the highest probability density,
               that is, see whether this is the winning distribution */
            mpfr_log(aux,DistributionProbDens,GMP_RNDN);
            MyLogDensity=mpfr_get_d(aux,GMP_RNDN);
            if (MyLogDensity>(*LogDensityProb))
            {
                (*LogDensityProb)=MyLogDensity;
                (*NdxWinRow)=NdxRowDist;
                (*NdxWinCol)=NdxColDist;
                /* Update the neuron responsibilities for the winning distribution */
                Index=0;
                for(NdxCol=0;NdxCol<NumColsMap;NdxCol++)
                {
                    for(NdxRow=0;NdxRow<NumRowsMap;NdxRow++)
                    {
                        ptrResponsibilities[Index]=NeighbourhoodFunction[Index]/SumNeighbourhoodFunctions;
                        Index++;
                    }
                }  
                
            }   
            /* End of the processing of this distribution */    

        }
    }  
    
    
    /* Release memory */
    mpfr_clear(DistributionProbDens);
    mpfr_clear(aux);
    mpfr_clear(MySumNeighbourhoodFunctions);
    for(Index=0;Index<NumColsMap*NumRowsMap;Index++)
    {
        mpfr_clear(ptrDensities[Index]);        
    } 
    
    mxFree(ptrVectorDif);
    mxFree(ptrVectorProd);
    mxFree(ptrDensities);
    mxFree(NeighbourhoodFunction);
    mxFree(ptrMyEigenValues);
    mxFree(ptrMyEigenVectors);
    
}    


/* Nonlinear function for Peel's method */
double FunctionPeel(double Nu,double AntAverage,double D,double DistMahal,
       double OldCoef,double NewCoef)
{
       double Factor,DistSigma,U,ForAverage,NewAverage;
       
       Factor=Nu/(Nu-2.0);
       DistSigma=Factor*DistMahal;
       U=(Nu+D)/(Nu+DistSigma);
       ForAverage=log(U)-U;
       NewAverage=OldCoef*AntAverage+NewCoef*ForAverage; 
                       
       return ( -psi(0.5*Nu)+log(0.5*Nu)+1.0+
            NewAverage+psi(0.5*(Nu+D))-log(0.5*(Nu+D)) );
}

/* Derivative of the nonlinear function for Peel's method */
double DerivFunctionPeel(double Nu,double AntAverage,double D,double DistMahal,
       double OldCoef,double NewCoef)
{
       double InvU,DerivU;
       
       InvU=(Nu+(Nu/(Nu-2.0))*DistMahal)/(Nu+D);
       DerivU=(Nu*Nu*DistMahal-D*Nu*Nu+4.0*D*Nu-4.0*D+2.0*D*DistMahal)/
            ((Nu-2.0+DistMahal)*(Nu-2.0+DistMahal)*Nu*Nu); 
                       
       return ( -0.5*trigamma(0.5*Nu)+1.0/Nu
              +NewCoef*(InvU-1.0)*DerivU
              +0.5*trigamma(0.5*(Nu+D))-1.0/(Nu+D) );
}
/* Nonlinear function for Shoham's method */
double FunctionShoham(double Nu,double AntAverage,double D,double DistMahal,
       double OldCoef,double NewCoef)
{       
       double Factor,DistSigma,L,U,ForAverage,NewAverage;
       
       Factor=Nu/(Nu-2.0);
       DistSigma=Factor*DistMahal;
       L=psi(0.5*(Nu+D))+
            log(2.0/(Nu+DistSigma));
       U=(Nu+D) / (Nu+DistSigma);
       ForAverage=L-U;
       NewAverage=OldCoef*AntAverage+NewCoef*ForAverage; 
       
       return ( NewAverage+
            log(0.5*Nu)+1.0-psi(0.5*Nu) );
}

/* Derivative of the nonlinear function for Shoham's method */
double DerivFunctionShoham(double Nu,double AntAverage,double D,double DistMahal,
       double OldCoef,double NewCoef)
{       
       double DerivTerminoL,DerivU;
       

       DerivTerminoL=-(Nu*Nu-4.0*Nu+4.0-2.0*DistMahal)/
            ((Nu-2.0)*Nu*(Nu-2.0+DistMahal));
          
       DerivU=(Nu*Nu*DistMahal-D*Nu*Nu+4.0*D*Nu-4.0*D+2.0*D*DistMahal)/
            ((Nu-2.0+DistMahal)*(Nu-2.0+DistMahal)*Nu*Nu); 

       return ( NewCoef*( 0.5*trigamma(0.5*(Nu+D))+DerivTerminoL-DerivU ) 
            +1.0/Nu-0.5*trigamma(0.5*Nu) );
}

/* Nonlinear function for the closed-form method */
double FunctionClosed(double Nu,double AntAverage,double D,double DistMahal,
       double OldCoef,double NewCoef)
{       
       double ForAverage,NewAverage;
       
       ForAverage=log(sqrt(DistMahal));
       NewAverage=OldCoef*AntAverage+NewCoef*ForAverage; 
       
       return ( 0.5*(psi(0.5*D)-psi(0.5*Nu)-log(1.0/(Nu-2.0)))-NewAverage );
}


/* Derivative of the nonlinear function for the closed-form method */
double DerivFunctionClosed(double Nu,double AntAverage,double D,double DistMahal,
       double OldCoef,double NewCoef)
{       
       
       return ( -0.25*trigamma(0.5*Nu)+0.5/(Nu-2.0) );
}

/* Solve a nonlinear equation in nu*/
#define MIN_NU 2.01
#define MAX_NU 10000.0
void SolveNuEquation(const mxArray* prhs[],double DistMahal,double D,
       double OldCoef,double NewCoef,double *ptrMyNu,double *ptrMyAverageNu) 
{
       int NumSections,Found,Method;
       double Start,Final,AntAverage,Factor,DistSigma,U,L,ForAverage,Nu;
       double IntervalStart,IntervalEnd,Delta,MyErrorAbs,BestErrorAbs;
       double (*FunctionNonLinear)(double,double,double,double,double,double); 
       double (*DerivFunctionNonLinear)(double,double,double,double,double,double); 
       double *ptrMethod;
       
       /* See which method we are using */
       ptrMethod=mxGetPr(prhs[2]);
       Method=(int)(*ptrMethod);
       switch(Method)
       {
            case METHOD_PEEL:
                 FunctionNonLinear=FunctionPeel;
                 DerivFunctionNonLinear=DerivFunctionPeel;
                 break;
            case METHOD_SHOHAM:
                 FunctionNonLinear=FunctionShoham;
                 DerivFunctionNonLinear=DerivFunctionShoham;                 
                 break;
            case METHOD_CLOSED:
                 FunctionNonLinear=FunctionClosed;
                 DerivFunctionNonLinear=DerivFunctionClosed;                 
                 break;
       }    
       
       AntAverage=(*ptrMyAverageNu);
       
                
       /* Search for an interval which contains the solution */
       NumSections=10;
       Found=0;
       Delta=(*ptrMyNu)/50.0;
       IntervalStart=(*ptrMyNu)-Delta;
       if (IntervalStart<MIN_NU)
       {
            IntervalStart=MIN_NU;
       }
       IntervalEnd=(*ptrMyNu)+Delta;
       if (IntervalEnd>MAX_NU)
       {
            IntervalEnd=MAX_NU;
       }
       
       while ((!Found) && ((IntervalStart>MIN_NU) || (IntervalEnd<MAX_NU)) && (NumSections<100) )
       {
               Found=FindSubinterval(FunctionNonLinear,AntAverage,D,DistMahal,OldCoef,NewCoef,
                  IntervalStart,IntervalEnd,NumSections,&Start,&Final);
               /* Enlarge the interval and increase the number of sections */
               Delta*=2.0;
               IntervalStart=(*ptrMyNu)-Delta;
               if (IntervalStart<MIN_NU)
               {
                    IntervalStart=MIN_NU;
               }
               IntervalEnd=(*ptrMyNu)+Delta;
               if (IntervalEnd>MAX_NU)
               {
                    IntervalEnd=MAX_NU;
               }
               NumSections++;
       }  
       
       /* If it has been impossible to find bounds for the solution, look for the value
       which yields the least absolute error.
       If the solution is bounded, find it and update AverageNu and Nu */
       if (Found)
       {            
               /* Find the solution in the specified interval */
               (*ptrMyNu)=FindSolution(FunctionNonLinear, DerivFunctionNonLinear, 
                    AntAverage, D, DistMahal, OldCoef, NewCoef,
                    Start, Final, 0.0001);
       }             
       else
       {
           Delta=(MAX_NU-MIN_NU)/100.0;
           BestErrorAbs=fabs(FunctionNonLinear(MIN_NU,AntAverage,D,DistMahal,OldCoef,NewCoef));
           (*ptrMyNu)=MIN_NU;
           
           for(Start=MIN_NU+Delta;Start<MAX_NU;Start+=Delta)
           {
                MyErrorAbs=fabs(FunctionNonLinear(Start,AntAverage,D,DistMahal,OldCoef,NewCoef));
                if (MyErrorAbs<BestErrorAbs)
                {
                     (*ptrMyNu)=Start;
                     BestErrorAbs=MyErrorAbs;
                }
           }
           
       }
       
       /* Update the average */
       Nu=(*ptrMyNu);
       switch(Method)
       {
            case METHOD_PEEL:
                 Factor=Nu/(Nu-2.0);
                 DistSigma=Factor*DistMahal;
                 U=(Nu+D)/(Nu+DistSigma);
                 ForAverage=log(U)-U;
                 (*ptrMyAverageNu)=OldCoef*AntAverage+NewCoef*ForAverage; 
                 break;
            case METHOD_SHOHAM:
                 Factor=Nu/(Nu-2.0);
                 DistSigma=Factor*DistMahal;
                 L=psi(0.5*(Nu+D))+
                      log(2.0/(Nu+DistSigma));
                 U=(Nu+D) / (Nu+DistSigma);
                 ForAverage=L-U;
                 (*ptrMyAverageNu)=OldCoef*AntAverage+NewCoef*ForAverage; 
                 break;
            case METHOD_CLOSED:
                 (*ptrMyAverageNu)=OldCoef*AntAverage
                         +NewCoef*log(sqrt(DistMahal));
                 break;
       } 

}     

/* Find an interval which contains a solution by splitting the interval [Start,End]
   in NumSlices subintervals */
int FindSubinterval(double (*NonlinearFunction)(double,double,double,double,double,double), double NewAverage, double D,
     double DistMahal,double OldCoef,double NewCoef,
     double Start, double End, int NumSlices,double *ResultStart, double *ResultEnd)
{
	int cnt;
	double x,fval1,fval2,xdelta,antx;
	
	
	
	xdelta=(End-Start)/NumSlices;
	x=Start;
	fval1=NonlinearFunction(x,NewAverage,D,DistMahal,OldCoef,NewCoef);
	for (cnt=0;cnt<NumSlices;cnt++)
    {
        antx=x;
        x+=xdelta;
		fval2=NonlinearFunction(x,NewAverage,D,DistMahal,OldCoef,NewCoef);
		if (( (fval2 > 0.0) && (fval1 < 0.0)) || ( (fval2 < 0.0) && (fval1 > 0.0)))
        {
			(*ResultStart)=antx;
			(*ResultEnd)=x;
			return -1;
		}
		fval1=fval2;
	}
	return 0;
}


/*
 ************************************************************************
 * Netlib repository: c/brent.shar
 *	    		    C math library
 * function ZEROIN - obtain a function zero within the given range
 *
 * Input
 *	double zeroin(ax,bx,f,tol)
 *	double ax; 			Root will be seeked for within
 *	double bx;  			a range [ax,bx]
 *	double (*f)(double x);		Name of the function whose zero
 *					will be seeked for
 *	double tol;			Acceptable tolerance for the root
 *					value.
 *					May be specified as 0.0 to cause
 *					the program to find the root as
 *					accurate as possible
 *
 * Output
 *	Zeroin returns an estimate for the root with accuracy
 *	4*EPSILON*abs(x) + tol
 *
 * Algorithm
 *	G.Forsythe, M.Malcolm, C.Moler, Computer methods for mathematical
 *	computations. M., Mir, 1980, p.180 of the Russian edition
 *
 *	The function makes use of the bissection procedure combined with
 *	the linear or quadric inverse interpolation.
 *	At every step program operates on three abscissae - a, b, and c.
 *	b - the last and the best approximation to the root
 *	a - the last but one approximation
 *	c - the last but one or even earlier approximation than a that
 *		1) |f(b)| <= |f(c)|
 *		2) f(b) and f(c) have opposite signs, i.e. b and c confine
 *		   the root
 *	At every step Zeroin selects one of the two new approximations, the
 *	former being obtained by the bissection procedure and the latter
 *	resulting in the interpolation (if a,b, and c are all different
 *	the quadric interpolation is utilized, otherwise the linear one).
 *	If the latter (i.e. obtained by the interpolation) point is 
 *	reasonable (i.e. lies within the current interval [b,c] not being
 *	too close to the boundaries) it is accepted. The bissection result
 *	is used in the other case. Therefore, the range of uncertainty is
 *	ensured to be reduced at least by the factor 1.6
 *
 ************************************************************************
 */

double zeroin(ax,bx,f,tol,NPr,D,DistMahal,OldCoef,NewCoef)		/* An estimate to the root	*/
double ax;				/* Left border | of the range	*/
double bx;  				/* Right border| the root is seeked*/
double (*f)(double,double,double,double,double,double);			/* Function under investigation	*/
double tol;				/* Acceptable tolerance		*/
double NPr;
double D;
double DistMahal;
double OldCoef;
double NewCoef;
{
  double a,b,c;				/* Abscissae, descr. see above	*/
  double fa;				/* f(a)				*/
  double fb;				/* f(b)				*/
  double fc;				/* f(c)				*/

  a = ax;  b = bx;  fa = (*f)(a,NPr,D,DistMahal,OldCoef,NewCoef);  
  fb = (*f)(b,NPr,D,DistMahal,OldCoef,NewCoef);
  c = a;   fc = fa;

  for(;;)		/* Main iteration loop	*/
  {
    double prev_step = b-a;		/* Distance from the last but one*/
					/* to the last approximation	*/
    double tol_act;			/* Actual tolerance		*/
    double p;      			/* Interpolation step is calcu- */
    double q;      			/* lated in the form p/q; divi- */
  					/* sion operations is delayed   */
 					/* until the last moment	*/
    double new_step;      		/* Step at this iteration       */
   
    if( fabs(fc) < fabs(fb) )
    {                         		/* Swap data for b to be the 	*/
	a = b;  b = c;  c = a;          /* best approximation		*/
	fa=fb;  fb=fc;  fc=fa;
    }
    tol_act = 2*DBL_EPSILON*fabs(b) + tol/2;
    new_step = (c-b)/2;

    if( fabs(new_step) <= tol_act || fb == (double)0 )
      return b;				/* Acceptable approx. is found	*/

    			/* Decide if the interpolation can be tried	*/
    if( fabs(prev_step) >= tol_act	/* If prev_step was large enough*/
	&& fabs(fa) > fabs(fb) )	/* and was in true direction,	*/
    {					/* Interpolatiom may be tried	*/
	register double t1,cb,t2;
	cb = c-b;
	if( a==c )			/* If we have only two distinct	*/
	{				/* points linear interpolation 	*/
	  t1 = fb/fa;			/* can only be applied		*/
	  p = cb*t1;
	  q = 1.0 - t1;
 	}
	else				/* Quadric inverse interpolation*/
	{
	  q = fa/fc;  t1 = fb/fc;  t2 = fb/fa;
	  p = t2 * ( cb*q*(q-t1) - (b-a)*(t1-1.0) );
	  q = (q-1.0) * (t1-1.0) * (t2-1.0);
	}
	if( p>(double)0 )		/* p was calculated with the op-*/
	  q = -q;			/* posite sign; make p positive	*/
	else				/* and assign possible minus to	*/
	  p = -p;			/* q				*/

	if( p < (0.75*cb*q-fabs(tol_act*q)/2)	/* If b+p/q falls in [b,c]*/
	    && p < fabs(prev_step*q/2) )	/* and isn't too large	*/
	  new_step = p/q;			/* it is accepted	*/
					/* If p/q is too large then the	*/
					/* bissection procedure can 	*/
					/* reduce [b,c] range to more	*/
					/* extent			*/
    }

    if( fabs(new_step) < tol_act )	/* Adjust the step to be not less*/
      if( new_step > (double)0 )	/* than tolerance		*/
	new_step = tol_act;
      else
	new_step = -tol_act;

    a = b;  fa = fb;			/* Save the previous approx.	*/
    b += new_step;  fb = (*f)(b,NPr,D,DistMahal,OldCoef,NewCoef);	/* Do step to a new approxim.	*/
    if( (fb > 0 && fc > 0) || (fb < 0 && fc < 0) )
    {                 			/* Adjust c for it to have a sign*/
      c = a;  fc = fa;                  /* opposite to that of b	*/
    }
  }

}

/* Find a solution of an equation. Derivative not
needed for Brent's method */
double FindSolution(double (*func)(double,double,double,double,double,double),
       double (*funcDeriv)(double,double,double,double,double,double),
       double NPr, double D,double DistMahal,double OldCoef,double NewCoef,
       const double Start, const double End,const double Accuracy)
{
	return zeroin(Start,End,func,Accuracy,NPr,D,DistMahal,OldCoef,NewCoef);
}

/*

  Purpose:

    ALNGAM computes the logarithm of the gamma function.

  Modified:

    13 January 2008

  Author:

    Allan Macleod
    C++ version by John Burkardt

  Reference:

    Allan Macleod,
    Algorithm AS 245,
    A Robust and Reliable Algorithm for the Logarithm of the Gamma Function,
    Applied Statistics,
    Volume 38, Number 2, 1989, pages 397-402.

  Parameters:

    Input, double XVALUE, the argument of the Gamma function.

    Output, int IFAULT, error flag.
    0, no error occurred.
    1, XVALUE is less than or equal to 0.
    2, XVALUE is too big.

    Output, double ALNGAM, the logarithm of the gamma function of X.
*/
double alngam ( double xvalue, int *ifault )
{
  double alr2pi = 0.918938533204673;
  double r1[9] = {
    -2.66685511495, 
    -24.4387534237, 
    -21.9698958928, 
     11.1667541262, 
     3.13060547623, 
     0.607771387771, 
     11.9400905721, 
     31.4690115749, 
     15.2346874070 };
  double r2[9] = {
    -78.3359299449, 
    -142.046296688, 
     137.519416416, 
     78.6994924154, 
     4.16438922228, 
     47.0668766060, 
     313.399215894, 
     263.505074721, 
     43.3400022514 };
  double r3[9] = {
    -2.12159572323E+05, 
     2.30661510616E+05, 
     2.74647644705E+04, 
    -4.02621119975E+04, 
    -2.29660729780E+03, 
    -1.16328495004E+05, 
    -1.46025937511E+05, 
    -2.42357409629E+04, 
    -5.70691009324E+02 };
  double r4[5] = {
     0.279195317918525, 
     0.4917317610505968, 
     0.0692910599291889, 
     3.350343815022304, 
     6.012459259764103 };
  double value;
  double x;
  double x1;
  double x2;
  double xlge = 510000.0;
  double xlgst = 1.0E+30;
  double y;

  x = xvalue;
  value = 0.0;
/*
  Check the input.
*/
  if ( xlgst <= x )
  {
    *ifault = 2;
    return value;
  }

  if ( x <= 0.0 )
  {
    *ifault = 1;
    return value;
  }

  *ifault = 0;
/*
  Calculation for 0 < X < 0.5 and 0.5 <= X < 1.5 combined.
*/
  if ( x < 1.5 )
  {
    if ( x < 0.5 )
    {
      value = - log ( x );
      y = x + 1.0;
/*
  Test whether X < machine epsilon.
*/
      if ( y == 1.0 )
      {
        return value;
      }
    }
    else
    {
      value = 0.0;
      y = x;
      x = ( x - 0.5 ) - 0.5;
    }

    value = value + x * (((( 
        r1[4]   * y 
      + r1[3] ) * y 
      + r1[2] ) * y 
      + r1[1] ) * y 
      + r1[0] ) / (((( 
                  y 
      + r1[8] ) * y 
      + r1[7] ) * y 
      + r1[6] ) * y 
      + r1[5] );

    return value;
  }
/*
  Calculation for 1.5 <= X < 4.0.
*/
  if ( x < 4.0 )
  {
    y = ( x - 1.0 ) - 1.0;

    value = y * (((( 
        r2[4]   * x 
      + r2[3] ) * x 
      + r2[2] ) * x 
      + r2[1] ) * x 
      + r2[0] ) / (((( 
                  x 
      + r2[8] ) * x 
      + r2[7] ) * x 
      + r2[6] ) * x 
      + r2[5] );
  }
/*
  Calculation for 4.0 <= X < 12.0.
*/
  else if ( x < 12.0 ) 
  {
    value = (((( 
        r3[4]   * x 
      + r3[3] ) * x 
      + r3[2] ) * x 
      + r3[1] ) * x 
      + r3[0] ) / (((( 
                  x 
      + r3[8] ) * x 
      + r3[7] ) * x 
      + r3[6] ) * x 
      + r3[5] );
  }
/*
  Calculation for 12.0 <= X.
*/
  else
  {
    y = log ( x );
    value = x * ( y - 1.0 ) - 0.5 * y + alr2pi;

    if ( x <= xlge )
    {
      x1 = 1.0 / x;
      x2 = x1 * x1;

      value = value + x1 * ( ( 
             r4[2]   * 
        x2 + r4[1] ) * 
        x2 + r4[0] ) / ( ( 
        x2 + r4[4] ) * 
        x2 + r4[3] );
    }
  }

  return value;
}

/* Natural logarithm of the gamma function */
double gammln(double argument)
{
	int ifault;
    
    return alngam(argument,&ifault);

}

/* psi.c -- psi function (i.e., the digamma function) for real arguments.
      Algorithms and coefficient values from "Computation of Special
      Functions", Zhang and Jin, John Wiley and Sons, 1996.

  (C) 2003, C. Bond. All rights reserved.
 http://www.crbond.com/math.htm

  Returns psi function for real argument 'x'.
  NOTE: Returns 1e308 for argument 0 or a negative integer.
*/

#define el 0.5772156649015329

double psi(double x)
{
    double s,ps,xa,End;
    int n,k;
    static double a[] = {
        -0.8333333333333e-01,
         0.83333333333333333e-02,
        -0.39682539682539683e-02,
         0.41666666666666667e-02,
        -0.75757575757575758e-02,
         0.21092796092796093e-01,
        -0.83333333333333333e-01,
         0.4432598039215686};

    xa = fabs(x);
    s = 0.0;
    if ((x == (int)x) && (x <= 0.0)) {
        ps = 1e308;
        return ps;
    }
    if (xa == (int)xa) {
        n = xa;
        for (k=1;k<n;k++) {
            s += 1.0/k;
        }
        ps =  s-el;
    }
    else if ((xa+0.5) == ((int)(xa+0.5))) {
        n = xa-0.5;
        for (k=1;k<=n;k++) {
            s += 1.0/(2.0*k-1.0);
        }
        ps = 2.0*s-el-1.386294361119891;
    }
    else {
        if (xa < 10.0) {
            n = 10-(int)xa;
            for (k=0;k<n;k++) {
                s += 1.0/(xa+k);
            }
            xa += n;
        }
        End = 1.0/(xa*xa);
        ps = log(xa)-0.5/xa+End*(((((((a[7]*End+a[6])*End+a[5])*End+
            a[4])*End+a[3])*End+a[2])*End+a[1])*End+a[0]);
        ps -= s;
    }
    if (x < 0.0)
        ps = ps - M_PI*cos(M_PI*x)/sin(M_PI*x)-1.0/x;
        return ps;
}

double trigamma ( double x )

/*

  Purpose:

    TRIGAMMA calculates the TriGamma function.

  Discussion:

    TriGamma(x) = d**2 log ( Gamma ( x ) ) / dx**2.

  Modified:

    03 January 2000

  Author:

    B Schneider

    C++ translation by John Burkardt

  Reference:

    B Schneider,
    Trigamma Function,
    Algorithm AS 121,
    Applied Statistics, 
    Volume 27, Number 1, page 97-99, 1978.

  Parameters:

    Input, double X, the argument of the trigamma function.
    0 < X.

    Output, double TRIGAMMA, the value of the 
    trigamma function at X.
*/
{
  double a = 0.0001;
  double b = 5.0;
  double b2 =   1.0 / 6.0;
  double b4 = - 1.0 / 30.0;
  double b6 =   1.0 / 42.0;
  double b8 = - 1.0 / 30.0;
  double value;
  double y;
  double z;
/*
  1): If X is not positive, fail.
*/
  if ( x <= 0.0 )
  {
    value = 0.0;
    mexErrMsgTxt("Trigamma: non positive argument");
  }
/*
  2): If X is smaller than A, use a small value approximation.
*/
  else if ( x <= a )
  {
    value = 1.0 / x / x;
  }
/*
  3): Otherwise, increase the argument to B <= ( X + I ).
*/
  else
  {
    z = x;
    value = 0.0;

    while ( z < b )
    {
      value = value + 1.0 / z / z;
      z = z + 1.0;
    }
/*
  ...and then apply an asymptotic formula.
*/
    y = 1.0 / z / z;

    value = value + 0.5 * 
            y + ( 1.0 
          + y * ( b2 
          + y * ( b4 
          + y * ( b6 
          + y *   b8 )))) / z;
  }

  return value;
}
